@@ -6,8 +6,8 @@
 #endif
 #include <assert.h>
 #include <limits.h>
-#include <msgpack/pack.h>
 #include <stdbool.h>
+#include <stddef.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -20,6 +20,7 @@
 #endif
 
 #include "auto/config.h"  // IWYU pragma: keep
+#include "klib/kvec.h"
 #include "nvim/api/extmark.h"
 #include "nvim/api/private/defs.h"
 #include "nvim/api/private/helpers.h"
@@ -37,13 +38,14 @@
 #include "nvim/diff.h"
 #include "nvim/drawline.h"
 #include "nvim/drawscreen.h"
+#include "nvim/errors.h"
 #include "nvim/eval.h"
 #include "nvim/eval/typval.h"
 #include "nvim/eval/typval_defs.h"
 #include "nvim/eval/userfunc.h"
 #include "nvim/event/loop.h"
 #include "nvim/event/multiqueue.h"
-#include "nvim/event/process.h"
+#include "nvim/event/proc.h"
 #include "nvim/event/stream.h"
 #include "nvim/ex_cmds.h"
 #include "nvim/ex_docmd.h"
@@ -79,9 +81,6 @@
 #include "nvim/option.h"
 #include "nvim/option_defs.h"
 #include "nvim/option_vars.h"
-#include "nvim/optionstr.h"
-#include "nvim/os/fileio.h"
-#include "nvim/os/fileio_defs.h"
 #include "nvim/os/fs.h"
 #include "nvim/os/input.h"
 #include "nvim/os/lang.h"
@@ -111,6 +110,9 @@
 
 #ifdef MSWIN
 # include "nvim/os/os_win_console.h"
+# ifndef _UCRT
+#  error UCRT is the only supported C runtime on windows
+# endif
 #endif
 
 #if defined(MSWIN) && !defined(MAKE_LIB)
@@ -154,7 +156,6 @@ void event_init(void)
   loop_init(&main_loop, NULL);
   resize_events = multiqueue_new_child(main_loop.events);
 
-  input_init();
   signal_init();
   // mspgack-rpc initialization
   channel_init();
@@ -164,7 +165,7 @@ void event_init(void)
 }
 
 /// @returns false if main_loop could not be closed gracefully
-bool event_teardown(void)
+static bool event_teardown(void)
 {
   if (!main_loop.events) {
     input_stop();
@@ -175,7 +176,7 @@ bool event_teardown(void)
   loop_poll_events(&main_loop, 0);  // Drain thread_events, fast_events.
   input_stop();
   channel_teardown();
-  process_teardown(&main_loop);
+  proc_teardown(&main_loop);
   timer_teardown();
   server_teardown();
   signal_teardown();
@@ -242,22 +243,10 @@ void early_init(mparm_T *paramp)
 #ifdef MAKE_LIB
 int nvim_main(int argc, char **argv);  // silence -Wmissing-prototypes
 int nvim_main(int argc, char **argv)
-#elif defined(MSWIN)
-int wmain(int argc, wchar_t **argv_w)  // multibyte args on Windows. #7060
 #else
 int main(int argc, char **argv)
 #endif
 {
-#if defined(MSWIN) && !defined(MAKE_LIB)
-  char **argv = xmalloc((size_t)argc * sizeof(char *));
-  for (int i = 0; i < argc; i++) {
-    char *buf = NULL;
-    utf16_to_utf8(argv_w[i], -1, &buf);
-    assert(buf);
-    argv[i] = buf;
-  }
-#endif
-
   argv0 = argv[0];
 
   if (!appname_is_valid()) {
@@ -267,7 +256,7 @@ int main(int argc, char **argv)
 
   if (argc > 1 && STRICMP(argv[1], "-ll") == 0) {
     if (argc == 2) {
-      print_mainerr(err_arg_missing, argv[1]);
+      print_mainerr(err_arg_missing, argv[1], NULL);
       exit(1);
     }
     nlua_run_script(argv, argc, 3);
@@ -333,12 +322,6 @@ int main(int argc, char **argv)
 #endif
   bool use_builtin_ui = (has_term && !headless_mode && !embedded_mode && !silent_mode);
 
-  // don't bind the server yet, if we are using builtin ui.
-  // This will be done when nvim server has been forked from the ui process
-  if (!use_builtin_ui) {
-    server_init(params.listen_addr);
-  }
-
   if (params.remote) {
     remote_request(&params, params.remote, params.server_addr, argc, argv,
                    use_builtin_ui);
@@ -356,11 +339,16 @@ int main(int argc, char **argv)
     ui_client_channel_id = rv;
   }
 
+  // NORETURN: Start builtin UI client.
   if (ui_client_channel_id) {
-    time_finish();
     ui_client_run(remote_ui);  // NORETURN
   }
   assert(!ui_client_channel_id && !use_builtin_ui);
+  // Nvim server...
+
+  if (!server_init(params.listen_addr)) {
+    mainerr(IObuff, NULL, NULL);
+  }
 
   TIME_MSG("expanding arguments");
 
@@ -372,7 +360,7 @@ int main(int argc, char **argv)
 
   setbuf(stdout, NULL);  // NOLINT(bugprone-unsafe-functions)
 
-  full_screen = !silent_mode || exmode_active;
+  full_screen = !silent_mode;
 
   // Set the default values for the options that use Rows and Columns.
   win_init_size();
@@ -639,13 +627,14 @@ int main(int argc, char **argv)
   }
 
   // WORKAROUND(mhi): #3023
-  if (cb_flags & CB_UNNAMEDMASK) {
+  if (cb_flags & (kOptCbFlagUnnamed | kOptCbFlagUnnamedplus)) {
     eval_has_provider("clipboard", false);
   }
 
   if (params.luaf != NULL) {
     // Like "--cmd", "+", "-c" and "-S", don't truncate messages.
     msg_scroll = true;
+    DLOG("executing Lua -l script");
     bool lua_ok = nlua_exec_file(params.luaf);
     TIME_MSG("executing Lua -l script");
     if (msg_didout) {
@@ -680,12 +669,14 @@ void os_exit(int r)
   } else {
     ui_flush();
     ui_call_stop();
-    ml_close_all(true);           // remove all memfiles
   }
 
   if (!event_teardown() && r == 0) {
     r = 1;  // Exit with error if main_loop did not teardown gracefully.
   }
+  if (!ui_client_channel_id) {
+    ml_close_all(true);  // remove all memfiles
+  }
   if (used_stdin) {
     stream_set_blocking(STDIN_FILENO, true);  // normalize stream (#2598)
   }
@@ -777,7 +768,11 @@ void getout(int exitval)
     }
   }
 
-  if (p_shada && *p_shada != NUL) {
+  if (
+#ifdef EXITFREE
+      !entered_free_all_mem &&
+#endif
+      p_shada && *p_shada != NUL) {
     // Write out the registers, history, marks etc, to the ShaDa file
     shada_write_file(NULL, false);
   }
@@ -850,8 +845,9 @@ void preserve_exit(const char *errmsg)
     // For TUI: exit alternate screen so that the error messages can be seen.
     ui_client_stop();
   }
-  if (errmsg != NULL) {
-    fprintf(stderr, "%s\n", errmsg);
+  if (errmsg != NULL && errmsg[0] != NUL) {
+    size_t has_eol = '\n' == errmsg[strlen(errmsg) - 1];
+    fprintf(stderr, has_eol ? "%s" : "%s\n", errmsg);
   }
   if (ui_client_channel_id) {
     os_exit(1);
@@ -862,7 +858,7 @@ void preserve_exit(const char *errmsg)
   FOR_ALL_BUFFERS(buf) {
     if (buf->b_ml.ml_mfp != NULL && buf->b_ml.ml_mfp->mf_fname != NULL) {
       if (errmsg != NULL) {
-        fprintf(stderr, "Vim: preserving files...\r\n");
+        fprintf(stderr, "Nvim: preserving files...\n");
       }
       ml_sync_all(false, false, true);  // preserve all swap files
       break;
@@ -872,7 +868,7 @@ void preserve_exit(const char *errmsg)
   ml_close_all(false);              // close all memfiles, without deleting
 
   if (errmsg != NULL) {
-    fprintf(stderr, "Vim: Finished.\r\n");
+    fprintf(stderr, "Nvim: Finished.\n");
   }
 
   getout(1);
@@ -969,8 +965,8 @@ static void remote_request(mparm_T *params, int remote_args, char *server_addr,
     os_exit(2);
   }
 
-  if (o.type == kObjectTypeDictionary) {
-    rvobj.data.dictionary = o.data.dictionary;
+  if (o.type == kObjectTypeDict) {
+    rvobj.data.dict = o.data.dict;
   } else {
     fprintf(stderr, "vim._cs_remote returned unexpected value\n");
     os_exit(2);
@@ -979,32 +975,32 @@ static void remote_request(mparm_T *params, int remote_args, char *server_addr,
   TriState should_exit = kNone;
   TriState tabbed = kNone;
 
-  for (size_t i = 0; i < rvobj.data.dictionary.size; i++) {
-    if (strequal(rvobj.data.dictionary.items[i].key.data, "errmsg")) {
-      if (rvobj.data.dictionary.items[i].value.type != kObjectTypeString) {
+  for (size_t i = 0; i < rvobj.data.dict.size; i++) {
+    if (strequal(rvobj.data.dict.items[i].key.data, "errmsg")) {
+      if (rvobj.data.dict.items[i].value.type != kObjectTypeString) {
         fprintf(stderr, "vim._cs_remote returned an unexpected type for 'errmsg'\n");
         os_exit(2);
       }
-      fprintf(stderr, "%s\n", rvobj.data.dictionary.items[i].value.data.string.data);
+      fprintf(stderr, "%s\n", rvobj.data.dict.items[i].value.data.string.data);
       os_exit(2);
-    } else if (strequal(rvobj.data.dictionary.items[i].key.data, "result")) {
-      if (rvobj.data.dictionary.items[i].value.type != kObjectTypeString) {
+    } else if (strequal(rvobj.data.dict.items[i].key.data, "result")) {
+      if (rvobj.data.dict.items[i].value.type != kObjectTypeString) {
         fprintf(stderr, "vim._cs_remote returned an unexpected type for 'result'\n");
         os_exit(2);
       }
-      printf("%s", rvobj.data.dictionary.items[i].value.data.string.data);
-    } else if (strequal(rvobj.data.dictionary.items[i].key.data, "tabbed")) {
-      if (rvobj.data.dictionary.items[i].value.type != kObjectTypeBoolean) {
+      printf("%s", rvobj.data.dict.items[i].value.data.string.data);
+    } else if (strequal(rvobj.data.dict.items[i].key.data, "tabbed")) {
+      if (rvobj.data.dict.items[i].value.type != kObjectTypeBoolean) {
         fprintf(stderr, "vim._cs_remote returned an unexpected type for 'tabbed'\n");
         os_exit(2);
       }
-      tabbed = rvobj.data.dictionary.items[i].value.data.boolean ? kTrue : kFalse;
-    } else if (strequal(rvobj.data.dictionary.items[i].key.data, "should_exit")) {
-      if (rvobj.data.dictionary.items[i].value.type != kObjectTypeBoolean) {
+      tabbed = rvobj.data.dict.items[i].value.data.boolean ? kTrue : kFalse;
+    } else if (strequal(rvobj.data.dict.items[i].key.data, "should_exit")) {
+      if (rvobj.data.dict.items[i].value.type != kObjectTypeBoolean) {
         fprintf(stderr, "vim._cs_remote returned an unexpected type for 'should_exit'\n");
         os_exit(2);
       }
-      should_exit = rvobj.data.dictionary.items[i].value.data.boolean ? kTrue : kFalse;
+      should_exit = rvobj.data.dict.items[i].value.data.boolean ? kTrue : kFalse;
     }
   }
   if (should_exit == kNone || tabbed == kNone) {
@@ -1052,7 +1048,7 @@ static void command_line_scan(mparm_T *parmp)
     // "+" or "+{number}" or "+/{pat}" or "+{command}" argument.
     if (argv[0][0] == '+' && !had_minmin) {
       if (parmp->n_commands >= MAX_ARG_CMDS) {
-        mainerr(err_extra_cmd, NULL);
+        mainerr(err_extra_cmd, NULL, NULL);
       }
       argv_idx = -1;  // skip to next argument
       if (argv[0][1] == NUL) {
@@ -1073,7 +1069,7 @@ static void command_line_scan(mparm_T *parmp)
           parmp->no_swap_file = true;
         } else {
           if (parmp->edit_type > EDIT_STDIN) {
-            mainerr(err_too_many_args, argv[0]);
+            mainerr(err_too_many_args, argv[0], NULL);
           }
           parmp->had_stdin_file = true;
           parmp->edit_type = EDIT_STDIN;
@@ -1098,23 +1094,13 @@ static void command_line_scan(mparm_T *parmp)
           // set stdout to binary to avoid crlf in --api-info output
           _setmode(STDOUT_FILENO, _O_BINARY);
 #endif
-          FileDescriptor fp;
-          const int fof_ret = file_open_fd(&fp, STDOUT_FILENO,
-                                           kFileWriteOnly);
-          if (fof_ret != 0) {
-            semsg(_("E5421: Failed to open stdin: %s"), os_strerror(fof_ret));
-          }
 
           String data = api_metadata_raw();
-          const ptrdiff_t written_bytes = file_write(&fp, data.data, data.size);
+          const ptrdiff_t written_bytes = os_write(STDOUT_FILENO, data.data, data.size, false);
           if (written_bytes < 0) {
-            msgpack_file_write_error((int)written_bytes);
+            semsg(_("E5420: Failed to write to file: %s"), os_strerror((int)written_bytes));
           }
 
-          const int ff_ret = file_flush(&fp);
-          if (ff_ret < 0) {
-            msgpack_file_write_error(ff_ret);
-          }
           os_exit(0);
         } else if (STRICMP(argv[0] + argv_idx, "headless") == 0) {
           headless_mode = true;
@@ -1146,7 +1132,7 @@ static void command_line_scan(mparm_T *parmp)
           nlua_disable_preload = true;
         } else {
           if (argv[0][argv_idx]) {
-            mainerr(err_opt_unknown, argv[0]);
+            mainerr(err_opt_unknown, argv[0], NULL);
           }
           had_minmin = true;
         }
@@ -1220,7 +1206,7 @@ static void command_line_scan(mparm_T *parmp)
         break;
       case 'q':    // "-q" QuickFix mode
         if (parmp->edit_type != EDIT_NONE) {
-          mainerr(err_too_many_args, argv[0]);
+          mainerr(err_too_many_args, argv[0], NULL);
         }
         parmp->edit_type = EDIT_QF;
         if (argv[0][argv_idx]) {  // "-q{errorfile}"
@@ -1243,13 +1229,16 @@ static void command_line_scan(mparm_T *parmp)
         if (exmode_active) {    // "-es" silent (batch) Ex-mode
           silent_mode = true;
           parmp->no_swap_file = true;
+          if (p_shadafile == NULL || *p_shadafile == NUL) {
+            set_option_value_give_err(kOptShadafile, STATIC_CSTR_AS_OPTVAL("NONE"), 0);
+          }
         } else {                // "-s {scriptin}" read from script file
           want_argument = true;
         }
         break;
       case 't':    // "-t {tag}" or "-t{tag}" jump to tag
         if (parmp->edit_type != EDIT_NONE) {
-          mainerr(err_too_many_args, argv[0]);
+          mainerr(err_too_many_args, argv[0], NULL);
         }
         parmp->edit_type = EDIT_TAG;
         if (argv[0][argv_idx]) {  // "-t{tag}"
@@ -1283,7 +1272,7 @@ static void command_line_scan(mparm_T *parmp)
       case 'c':    // "-c{command}" or "-c {command}" exec command
         if (argv[0][argv_idx] != NUL) {
           if (parmp->n_commands >= MAX_ARG_CMDS) {
-            mainerr(err_extra_cmd, NULL);
+            mainerr(err_extra_cmd, NULL, NULL);
           }
           parmp->commands[parmp->n_commands++] = argv[0] + argv_idx;
           argv_idx = -1;
@@ -1300,19 +1289,19 @@ static void command_line_scan(mparm_T *parmp)
         break;
 
       default:
-        mainerr(err_opt_unknown, argv[0]);
+        mainerr(err_opt_unknown, argv[0], NULL);
       }
 
       // Handle option arguments with argument.
       if (want_argument) {
         // Check for garbage immediately after the option letter.
         if (argv[0][argv_idx] != NUL) {
-          mainerr(err_opt_garbage, argv[0]);
+          mainerr(err_opt_garbage, argv[0], NULL);
         }
 
         argc--;
         if (argc < 1 && c != 'S') {  // -S has an optional argument
-          mainerr(err_arg_missing, argv[0]);
+          mainerr(err_arg_missing, argv[0], NULL);
         }
         argv++;
         argv_idx = -1;
@@ -1321,7 +1310,7 @@ static void command_line_scan(mparm_T *parmp)
         case 'c':    // "-c {command}" execute command
         case 'S':    // "-S {file}" execute Vim script
           if (parmp->n_commands >= MAX_ARG_CMDS) {
-            mainerr(err_extra_cmd, NULL);
+            mainerr(err_extra_cmd, NULL, NULL);
           }
           if (c == 'S') {
             char *a;
@@ -1352,7 +1341,7 @@ static void command_line_scan(mparm_T *parmp)
           if (strequal(argv[-1], "--cmd")) {
             // "--cmd {command}" execute command
             if (parmp->n_pre_commands >= MAX_ARG_CMDS) {
-              mainerr(err_extra_cmd, NULL);
+              mainerr(err_extra_cmd, NULL, NULL);
             }
             parmp->pre_commands[parmp->n_pre_commands++] = argv[0];
           } else if (strequal(argv[-1], "--listen")) {
@@ -1434,7 +1423,7 @@ scripterror:
 
       // Check for only one type of editing.
       if (parmp->edit_type > EDIT_STDIN) {
-        mainerr(err_too_many_args, argv[0]);
+        mainerr(err_too_many_args, argv[0], NULL);
       }
       parmp->edit_type = EDIT_FILE;
 
@@ -1442,6 +1431,17 @@ scripterror:
       ga_grow(&global_alist.al_ga, 1);
       char *p = xstrdup(argv[0]);
 
+      // On Windows expand "~\" or "~/" prefix in file names to profile directory.
+#ifdef MSWIN
+      if (*p == '~' && (p[1] == '\\' || p[1] == '/')) {
+        size_t size = strlen(os_homedir()) + strlen(p);
+        char *tilde_expanded = xmalloc(size);
+        snprintf(tilde_expanded, size, "%s%s", os_homedir(), p + 1);
+        xfree(p);
+        p = tilde_expanded;
+      }
+#endif
+
       if (parmp->diff_mode && os_isdir(p) && GARGCOUNT > 0
           && !os_isdir(alist_name(&GARGLIST[0]))) {
         char *r = concat_fnames(p, path_tail(alist_name(&GARGLIST[0])), true);
@@ -1449,7 +1449,7 @@ scripterror:
         p = r;
       }
 
-#ifdef USE_FNAME_CASE
+#ifdef CASE_INSENSITIVE_FILENAME
       // Make the case of the file name match the actual file.
       path_fix_case(p);
 #endif
@@ -1470,7 +1470,7 @@ scripterror:
   }
 
   if (embedded_mode && (silent_mode || parmp->luaf)) {
-    mainerr(_("--embed conflicts with -es/-Es/-l"), NULL);
+    mainerr(_("--embed conflicts with -es/-Es/-l"), NULL, NULL);
   }
 
   // If there is a "+123" or "-c" command, set v:swapcommand to the first one.
@@ -1514,7 +1514,7 @@ static void init_startuptime(mparm_T *paramp)
   }
   for (int i = 1; i < paramp->argc - 1; i++) {
     if (STRICMP(paramp->argv[i], "--startuptime") == 0) {
-      time_init(paramp->argv[i + 1], is_embed ? "Embedded" : "Primary/TUI");
+      time_init(paramp->argv[i + 1], is_embed ? "Embedded" : "Primary (or UI client)");
       time_start("--- NVIM STARTING ---");
       break;
     }
@@ -2064,7 +2064,7 @@ static void do_exrc_initialization(void)
       xfree(str);
       if (ERROR_SET(&err)) {
         semsg("Error detected while processing %s:", VIMRC_LUA_FILE);
-        semsg_multiline(err.msg);
+        semsg_multiline("emsg", err.msg);
         api_clear_error(&err);
       }
     }
@@ -2089,8 +2089,7 @@ static void source_startup_scripts(const mparm_T *const parmp)
 {
   // If -u given, use only the initializations from that file and nothing else.
   if (parmp->use_vimrc != NULL) {
-    if (strequal(parmp->use_vimrc, "NONE")
-        || strequal(parmp->use_vimrc, "NORC")) {
+    if (strequal(parmp->use_vimrc, "NONE") || strequal(parmp->use_vimrc, "NORC")) {
       // Do nothing.
     } else {
       if (do_source(parmp->use_vimrc, false, DOSO_NONE, NULL) != OK) {
@@ -2133,28 +2132,30 @@ static int execute_env(char *env)
   return OK;
 }
 
-/// Prints the following then exits:
-/// - An error message `errstr`
-/// - A string `str` if not null
+/// Prints a message of the form "{msg1}: {msg2}: {msg3}", then exits with code 1.
 ///
-/// @param errstr  string containing an error message
-/// @param str     string to append to the primary error message, or NULL
-static void mainerr(const char *errstr, const char *str)
+/// @param msg1  error message
+/// @param msg2  extra message, or NULL
+/// @param msg3  extra message, or NULL
+static void mainerr(const char *msg1, const char *msg2, const char *msg3)
   FUNC_ATTR_NORETURN
 {
-  print_mainerr(errstr, str);
+  print_mainerr(msg1, msg2, msg3);
   os_exit(1);
 }
 
-static void print_mainerr(const char *errstr, const char *str)
+static void print_mainerr(const char *msg1, const char *msg2, const char *msg3)
 {
   char *prgname = path_tail(argv0);
 
   signal_stop();              // kill us with CTRL-C here, if you like
 
-  fprintf(stderr, "%s: %s", prgname, _(errstr));
-  if (str != NULL) {
-    fprintf(stderr, ": \"%s\"", str);
+  fprintf(stderr, "%s: %s", prgname, _(msg1));
+  if (msg2 != NULL) {
+    fprintf(stderr, ": \"%s\"", msg2);
+  }
+  if (msg3 != NULL) {
+    fprintf(stderr, ": \"%s\"", msg3);
   }
   fprintf(stderr, _("\nMore info with \""));
   fprintf(stderr, "%s -h\"\n", prgname);
@@ -2205,7 +2206,7 @@ static void usage(void)
   printf(_("  --headless            Don't start a user interface\n"));
   printf(_("  --listen <address>    Serve RPC API from this address\n"));
   printf(_("  --remote[-subcommand] Execute commands remotely on a server\n"));
-  printf(_("  --server <address>    Specify RPC server to send commands to\n"));
+  printf(_("  --server <address>    Connect to this Nvim server\n"));
   printf(_("  --startuptime <file>  Write startup timing messages to <file>\n"));
   printf(_("\nSee \":help startup-options\" for all options.\n"));
 }
