@@ -886,7 +886,7 @@ local function test_cmdline(linegrid)
       },
       cmdline_block = { { { 'if 1' } } },
     })
-    feed(':let x = 1<CR>')
+    feed('let x = 1<CR>')
     screen:expect({
       grid = s1,
       cmdline = {
@@ -897,20 +897,34 @@ local function test_cmdline(linegrid)
           pos = 0,
         },
       },
-      cmdline_block = { { { 'if 1' } }, { { '  :let x = 1' } } },
+      cmdline_block = { { { 'if 1' } }, { { '  let x = 1' } } },
     })
-    feed(':endif')
+    feed('<CR>')
+    eq('let x = 1', eval('@:'))
     screen:expect({
       grid = s1,
       cmdline = {
         {
-          content = { { ':endif' } },
+          content = { { '' } },
           firstc = ':',
           indent = 2,
-          pos = 6,
+          pos = 0,
         },
       },
-      cmdline_block = { { { 'if 1' } }, { { '  :let x = 1' } } },
+      cmdline_block = { { { 'if 1' } }, { { '  let x = 1' } }, { { '  ' } } },
+    })
+    feed('endif')
+    screen:expect({
+      grid = s1,
+      cmdline = {
+        {
+          content = { { 'endif' } },
+          firstc = ':',
+          indent = 2,
+          pos = 5,
+        },
+      },
+      cmdline_block = { { { 'if 1' } }, { { '  let x = 1' } }, { { '  ' } } },
     })
     feed('<CR>')
     screen:expect({
@@ -1352,6 +1366,12 @@ describe('cmdline height', function()
     -- cmdheight unchanged.
     eq(1, eval('&cmdheight'))
   end)
+
+  it('not increased to 0 from 1 with wincmd _', function()
+    command('set cmdheight=0 laststatus=0')
+    command('wincmd _')
+    eq(0, eval('&cmdheight'))
+  end)
 end)
 
 describe('cmdheight=0', function()
